"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.installPluginEntry = void 0;
const fs_1 = __importDefault(require("fs"));
const koishi_1 = require("koishi");
const child_process_1 = require("child_process");
const util_1 = require("util");
const path_1 = __importDefault(require("path"));
const read_package_json_1 = require("./utils/read-package-json");
const loader_1 = __importDefault(require("@koishijs/loader"));
const execFileAsync = (0, util_1.promisify)(child_process_1.execFile);
const logger = new koishi_1.Logger('bootstrap-install-plugin');
async function checkPluginExists(name, allowCommunity, allowOfficial) {
    const deps = await (0, read_package_json_1.getPackageJsonPackages)();
    const entry = Object.entries(deps).find(([packageName, version]) => packageName === name ||
        (allowOfficial && packageName === `@koishijs/plugin-${name}`) ||
        (allowCommunity && packageName.endsWith(`koishi-plugin-${name}`)));
    if (!entry) {
        return;
    }
    try {
        const stat = await fs_1.default.promises.stat(path_1.default.join('node_modules', entry[0]));
        if (!stat?.isDirectory()) {
            return;
        }
        return entry[1];
    }
    catch (e) {
        return;
    }
}
async function npmInstall(name) {
    logger.info(`Installing package ${name}.`);
    try {
        await execFileAsync('npm', [
            'install',
            '--save-exact',
            '--loglevel',
            'error',
            name,
        ]);
        logger.info(`Package ${name} installed.`);
        return true;
    }
    catch (e) {
        logger.warn(`Package ${name} not found.`);
        return false;
    }
}
async function tryInstallPackages(names) {
    if (process.env.DRY_RUN) {
        logger.info(`Will install packages ${names.join(', ')}.`);
        return true;
    }
    for (const name of names) {
        if (await npmInstall(name)) {
            return true;
        }
    }
    return false;
}
async function installPlugin(name, info) {
    const suffixMatching = name.match(/(.+):.+/);
    if (suffixMatching) {
        name = suffixMatching[1];
    }
    if (name.match(/^([\.\/~\\]|[A-Za-z]:[\/\\])/)) {
        logger.info(`Plugin ${name} is a local plugin, skipping.`);
        return;
    }
    const version = info.$version ||
        (typeof info.$install === 'string' ? info.$install : undefined);
    logger.info(`Installing plugin ${name}@${version || '*'}.`);
    const allowCommunity = !info.$official;
    const allowOfficial = !info.$community;
    if (!allowCommunity && !allowOfficial) {
        logger.warn(`Plugin ${name} is neither official nor community, skipping.`);
        return;
    }
    const existingPluginVersion = await checkPluginExists(name, allowCommunity, allowOfficial);
    if (existingPluginVersion &&
        (existingPluginVersion === version || !version || version === '*')) {
        logger.info(`Plugin ${name}@${existingPluginVersion} exists, skipping.`);
        return;
    }
    const installList = [];
    const communityPrefix = 'koishi-plugin-';
    const officialPrefix = '@koishijs/plugin-';
    const installVersion = version || 'latest';
    if (name.includes(communityPrefix) || name.startsWith(officialPrefix)) {
        installList.push(`${name}@${installVersion}`);
    }
    else if (name.startsWith('@') && name.includes('/')) {
        const [author, pluginName] = name.split('/');
        const installName = pluginName.startsWith(communityPrefix)
            ? pluginName
            : `${communityPrefix}${pluginName}`;
        installList.push(`${author}/${communityPrefix}${installName}@${installVersion}`);
    }
    else {
        if (allowOfficial) {
            installList.push(`${officialPrefix}${name}@${installVersion}`);
        }
        if (allowCommunity) {
            installList.push(`${communityPrefix}${name}@${installVersion}`);
        }
    }
    if (!installList.length) {
        logger.warn(`Plugin ${name} has nothing to install.`);
        return;
    }
    const result = await tryInstallPackages(installList);
    if (!result) {
        logger.error(`Plugin ${name}@${version || 'unknown'} install failed.`);
    }
}
async function installPluginEntry() {
    logger.info(`Bootstrapping...`);
    const config = new loader_1.default().readConfig();
    const plugins = config?.plugins;
    if (!plugins) {
        logger.warn(`No plugins found, exiting.`);
        return;
    }
    // console.log(config.plugins);
    logger.info(`Cleaning NPM cache.`);
    await execFileAsync('npm', ['cache', 'clean', '--force']);
    let queue = Object.entries(plugins).reverse();
    while (queue.length) {
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        const [name, info] = queue.pop();
        if (name.startsWith('~') || !info) {
            // disabled plugin
            continue;
        }
        if (name.startsWith('group:')) {
            queue = queue.concat(Object.entries(info).reverse());
            continue;
        }
        if (!info.$install && !info.$version) {
            continue;
        }
        await installPlugin(name, info);
    }
    logger.info(`Bootstrap finished.`);
}
exports.installPluginEntry = installPluginEntry;
//# sourceMappingURL=install-plugin.js.map